=== added file '__init__.py'
--- __init__.py	1970-01-01 00:00:00 +0000
+++ __init__.py	2007-06-11 02:10:58 +0000
@@ -0,0 +1,174 @@
+"""Implementation of multiparent diffs for versionedfile-like storage
+
+Provides mp-regen and mp-extract commands.
+Focus is on comparing size/performance to knits.
+"""
+
+from bzrlib.lazy_import import lazy_import
+
+lazy_import(globals(), """
+import (
+        errno,
+        os,
+        sys,
+        time,
+        )
+
+from bzrlib import (
+    commands,
+    urlutils
+    )
+from bzrlib.workingtree import WorkingTree
+from bzrlib.tests import TestUtil
+
+from bzrlib.plugins.multiparent.multiparent import (
+    MultiVersionedFile,
+    MultiMemoryVersionedFile,
+    )
+""")
+
+class cmd_mp_regen(commands.Command):
+    """Generate a multiparent versionedfile"""
+
+    takes_args = ['file?']
+    takes_options = [commands.Option('sync-snapshots',
+                                     help='Snapshots follow source'),
+                     commands.Option('snapshot-interval', type=int,
+                                     help='take snapshots every x revisions'),
+                     commands.Option('outfile', type=unicode,
+                                     help='Write pseudo-knit to this file'),
+                     commands.Option('memory', help='Use memory, not disk'),
+                     commands.Option('extract', help='test extract time'),
+                     commands.Option('single', help='use a single parent'),
+                     commands.Option('verify', help='verify added texts'),
+                     commands.Option('cache', help='Aggresively cache'),
+                     commands.Option('size', help='Aggressive size'),
+                     commands.Option('build', help='Aggressive build'),
+                    ]
+    hidden = True
+
+    def run(self, file=None, sync_snapshots=False, snapshot_interval=26,
+            lsprof_timed=False, dump=False, extract=False, single=False,
+            verify=False, outfile=None, memory=False, cache=False,
+            size=False, build=False):
+        file_weave = get_file_weave(file)
+        url = file_weave.transport.abspath(file_weave.filename)
+        print >> sys.stderr, 'Importing: %s' % \
+            urlutils.local_path_from_url(url)
+        if sync_snapshots:
+            print >> sys.stderr, 'Snapshots follow input'
+        else:
+            print >> sys.stderr, 'Snapshot interval: %d' % snapshot_interval
+        if not memory:
+            if outfile is None:
+                filename = 'pknit'
+            else:
+                filename = outfile
+            vf = MultiVersionedFile(filename, snapshot_interval)
+        else:
+            vf = MultiMemoryVersionedFile(snapshot_interval)
+        vf.destroy()
+        old_snapshots = set(r for r in file_weave.versions() if
+                        file_weave._index.get_method(r) == 'fulltext')
+        if sync_snapshots:
+            to_sync = old_snapshots
+        elif size or build:
+            assert memory
+            to_sync = set()
+        else:
+            to_sync = vf.select_snapshots(file_weave)
+        print >> sys.stderr, "%d fulltext(s)" % len(old_snapshots)
+        print >> sys.stderr, "%d planned snapshots" % len(to_sync)
+
+        try:
+            vf.import_versionedfile(file_weave, to_sync, single_parent=single,
+                                    verify=verify, no_cache=not cache)
+            if size:
+                snapshots = vf.select_by_size(len(old_snapshots))
+                for version_id in snapshots:
+                    vf.make_snapshot(version_id)
+            if build:
+                ranking = vf.get_build_ranking()
+                snapshots = ranking[:len(old_snapshots) -\
+                    len(vf._snapshots)]
+                for version_id in snapshots:
+                    vf.make_snapshot(version_id)
+        except:
+            vf.destroy()
+            raise
+        try:
+            print >> sys.stderr, "%d actual snapshots" % len(vf._snapshots)
+            if not cache:
+                vf.clear_cache()
+            if memory:
+                if outfile is not None:
+                    vf_file = MultiVersionedFile(outfile)
+                vf_file.import_diffs(vf)
+            else:
+                vf_file = vf
+        finally:
+            if outfile is None:
+                vf.destroy()
+            else:
+                vf_file.save()
+
+class cmd_mp_extract(commands.Command):
+    """Test extraction time multiparent knits"""
+
+    takes_options = [
+        commands.Option('lsprof-timed', help='Use lsprof'),
+        commands.Option('parallel', help='extract multiple versions at once'),
+        commands.Option('count', help='Number of cycles to do', type=int),
+        ]
+
+    takes_args = ['filename', 'vfile?']
+
+    def run(self, filename, vfile=None, lsprof_timed=False, count=1000,
+            parallel=False):
+        vf = MultiVersionedFile(filename)
+        vf.load()
+        snapshots = [r for r in vf.versions() if vf.get_diff(r).is_snapshot()]
+        print '%d snapshots' % len(snapshots)
+        revisions = list(vf.versions())
+        revisions = revisions[-count:]
+        print 'Testing extract time of %d revisions' % len(revisions)
+        if parallel:
+            revisions_list = [revisions]
+        else:
+            revisions_list = [[r] for r in revisions]
+        start = time.clock()
+        for revisions in revisions_list:
+            vf = MultiVersionedFile(filename)
+            vf.load()
+            vf.get_line_list(revisions)
+        print >> sys.stderr, time.clock() - start
+        if lsprof_timed:
+            from bzrlib.lsprof import profile
+            vf.clear_cache()
+            ret, stats = profile(vf.get_line_list, revisions_list[-1][-1])
+            stats.sort()
+            stats.pprint()
+        start = time.clock()
+        for revisions in revisions_list:
+            file_weave = get_file_weave(vfile)
+            file_weave.get_line_list(revisions)
+        print >> sys.stderr, time.clock() - start
+
+
+def get_file_weave(filename=None, wt=None):
+    if filename is None:
+        wt, path = WorkingTree.open_containing('.')
+        return wt.branch.repository.get_inventory_weave()
+    else:
+        wt, path = WorkingTree.open_containing(filename)
+        file_id = wt.path2id(path)
+        bt = wt.branch.repository.revision_tree(wt.last_revision())
+        return bt.get_weave(file_id)
+
+
+commands.register_command(cmd_mp_regen)
+commands.register_command(cmd_mp_extract)
+
+def test_suite():
+    from bzrlib.plugins.multiparent import test_multiparent
+    return TestUtil.TestLoader().loadTestsFromModule(test_multiparent)

=== added file 'mpdiff'
--- mpdiff	1970-01-01 00:00:00 +0000
+++ mpdiff	2007-06-11 02:10:58 +0000
@@ -0,0 +1,8 @@
+#!/usr/bin/env python2.4
+import sys
+
+from multiparent import MultiParent
+
+text = open(sys.argv[1], 'rb').readlines()
+parents = [open(p, 'rb').readlines() for p in sys.argv[2:]]
+sys.stdout.writelines(MultiParent.from_lines(text, parents).to_patch())

=== added file 'mpknit'
--- mpknit	1970-01-01 00:00:00 +0000
+++ mpknit	2007-06-11 02:10:58 +0000
@@ -0,0 +1,31 @@
+#!/usr/bin/env python2.4
+import sys
+
+from bzrlib.workingtree import WorkingTree 
+
+from multiparent import MultiParent
+single_parent = False
+if len(sys.argv) > 1 and len(sys.argv) < 4:
+    wt, path = WorkingTree.open_containing(sys.argv[-1])
+    if len(sys.argv) == 3:
+        assert sys.argv[1] == '--single'
+        single_parent = True
+else:
+    print >> sys.stderr, 'Usage: mpknit [--single] FILENAME'
+    sys.exit(3)
+    
+wt.lock_read()
+try:
+    bt = wt.branch.repository.revision_tree(wt.last_revision())
+    file_id = wt.path2id(path)
+    file_weave = bt.get_weave(file_id)
+    for revision in file_weave.get_ancestry([bt.inventory[file_id].revision]):
+        parents = file_weave.get_parents(revision)
+        if single_parent:
+            parents = parents[0:1]
+        texts = file_weave.get_texts([revision] + parents)
+        print revision
+        patch_iter = MultiParent.from_texts(texts[0], texts[1:]).to_patch()
+        sys.stdout.writelines(patch_iter)
+finally:
+    wt.unlock()

=== added file 'multiparent.py'
--- multiparent.py	1970-01-01 00:00:00 +0000
+++ multiparent.py	2007-06-11 02:10:58 +0000
@@ -0,0 +1,571 @@
+from bzrlib.lazy_import import lazy_import
+
+lazy_import(globals(), """
+import errno
+import itertools
+import os
+from StringIO import StringIO
+
+from bzrlib import (
+    patiencediff,
+    trace,
+    ui,
+    )
+from bzrlib.util import bencode
+""")
+from bzrlib.tuned_gzip import GzipFile
+
+
+def topo_iter(vf):
+    seen = set()
+    descendants = {}
+    for version_id in vf.versions():
+        for parent_id in vf.get_parents(version_id):
+            descendants.setdefault(parent_id, []).append(version_id)
+    cur = [v for v in vf.versions() if len(vf.get_parents(v)) == 0]
+    while len(cur) > 0:
+        next = []
+        for version_id in cur:
+            if version_id in seen:
+                continue
+            parents = vf.get_parents(version_id)
+            if not seen.issuperset(parents):
+                continue
+            next.extend(descendants.get(version_id, []))
+            yield version_id
+            seen.add(version_id)
+        cur = next
+
+
+class MultiParent(object):
+
+    def __init__(self, hunks=None):
+        if hunks is not None:
+            self.hunks = hunks
+        else:
+            self.hunks = []
+
+    def __repr__(self):
+        return "MultiParent(%r)" % self.hunks
+
+    def __eq__(self, other):
+        if self.__class__ is not other.__class__:
+            return False
+        return (self.hunks == other.hunks)
+
+    @staticmethod
+    def from_lines(text, parents=()):
+        """Produce a MultiParent from a list of lines and parents"""
+        def compare(parent):
+            matcher = patiencediff.PatienceSequenceMatcher(None, parent,
+                                                           text)
+            return matcher.get_matching_blocks()
+        parent_comparisons = [compare(p) for p in parents]
+        cur_line = 0
+        new_text = NewText([])
+        parent_text = []
+        block_iter = [iter(i) for i in parent_comparisons]
+        diff = MultiParent([])
+        def next_block(p):
+            try:
+                return block_iter[p].next()
+            except StopIteration:
+                return None
+        cur_block = [next_block(p) for p, i in enumerate(block_iter)]
+        while cur_line < len(text):
+            best_match = None
+            for p, block in enumerate(cur_block):
+                if block is None:
+                    continue
+                i, j, n = block
+                while j + n < cur_line:
+                    block = cur_block[p] = next_block(p)
+                    if block is None:
+                        break
+                    i, j, n = block
+                if block is None:
+                    continue
+                if j > cur_line:
+                    continue
+                offset = cur_line - j
+                i += offset
+                j = cur_line
+                n -= offset
+                if n == 0:
+                    continue
+                if best_match is None or n > best_match.num_lines:
+                    best_match = ParentText(p, i, j, n)
+            if best_match is None:
+                new_text.lines.append(text[cur_line])
+                cur_line += 1
+            else:
+                if len(new_text.lines) > 0:
+                    diff.hunks.append(new_text)
+                    new_text = NewText([])
+                diff.hunks.append(best_match)
+                cur_line += best_match.num_lines
+        if len(new_text.lines) > 0:
+            diff.hunks.append(new_text)
+        return diff
+
+    @classmethod
+    def from_texts(cls, text, parents=()):
+        """Produce a MultiParent from a text and list of parent text"""
+        return cls.from_lines(text.splitlines(True),
+                              [p.splitlines(True) for p in parents])
+
+    def to_patch(self):
+        """Yield text lines for a patch"""
+        for hunk in self.hunks:
+            for line in hunk.to_patch():
+                yield line
+
+    def patch_len(self):
+        return len(''.join(self.to_patch()))
+
+    def zipped_patch_len(self):
+        return len(gzip_string(self.to_patch()))
+
+    @staticmethod
+    def from_patch(lines):
+        """Produce a MultiParent from a sequence of lines"""
+        line_iter = iter(lines)
+        hunks = []
+        cur_line = None
+        while(True):
+            try:
+                cur_line = line_iter.next()
+            except StopIteration:
+                break
+            if cur_line[0] == 'i':
+                num_lines = int(cur_line.split(' ')[1])
+                hunk_lines = [line_iter.next() for x in xrange(num_lines)]
+                hunk_lines[-1] = hunk_lines[-1][:-1]
+                hunks.append(NewText(hunk_lines))
+            elif cur_line[0] == '\n':
+                hunks[-1].lines[-1] += '\n'
+            else:
+                assert cur_line[0] == 'c', cur_line[0]
+                parent, parent_pos, child_pos, num_lines =\
+                    [int(v) for v in cur_line.split(' ')[1:]]
+                hunks.append(ParentText(parent, parent_pos, child_pos,
+                                        num_lines))
+        return MultiParent(hunks)
+
+    def range_iterator(self):
+        """Iterate through the hunks, with range indicated
+
+        kind is "new" or "parent".
+        for "new", data is a list of lines.
+        for "parent", data is (parent, parent_start, parent_end)
+        :return: a generator of (start, end, kind, data)
+        """
+        start = 0
+        for hunk in self.hunks:
+            if isinstance(hunk, NewText):
+                kind = 'new'
+                end = start + len(hunk.lines)
+                data = hunk.lines
+            else:
+                kind = 'parent'
+                start = hunk.child_pos
+                end = start + hunk.num_lines
+                data = (hunk.parent, hunk.parent_pos, hunk.parent_pos +
+                        hunk.num_lines)
+            yield start, end, kind, data
+            start = end
+
+    def num_lines(self):
+        extra_n = 0
+        for hunk in reversed(self.hunks):
+            if isinstance(hunk, ParentText):
+               return hunk.child_pos + hunk.num_lines + extra_n
+            extra_n += len(hunk.lines)
+        return extra_n
+
+    def is_snapshot(self):
+        if len(self.hunks) != 1:
+            return False
+        return (isinstance(self.hunks[0], NewText))
+
+
+class NewText(object):
+    """The contents of text that is introduced by this text"""
+
+    def __init__(self, lines):
+        self.lines = lines
+
+    def __eq__(self, other):
+        if self.__class__ is not other.__class__:
+            return False
+        return (other.lines == self.lines)
+
+    def __repr__(self):
+        return 'NewText(%r)' % self.lines
+
+    def to_patch(self):
+        yield 'i %d\n' % len(self.lines)
+        for line in self.lines:
+            yield line
+        yield '\n'
+
+
+class ParentText(object):
+    """A reference to text present in a parent text"""
+
+    def __init__(self, parent, parent_pos, child_pos, num_lines):
+        self.parent = parent
+        self.parent_pos = parent_pos
+        self.child_pos = child_pos
+        self.num_lines = num_lines
+
+    def __repr__(self):
+        return 'ParentText(%(parent)r, %(parent_pos)r, %(child_pos)r,'\
+            ' %(num_lines)r)' % self.__dict__
+
+    def __eq__(self, other):
+        if self.__class__ != other.__class__:
+            return False
+        return (self.__dict__ == other.__dict__)
+
+    def to_patch(self):
+        yield 'c %(parent)d %(parent_pos)d %(child_pos)d %(num_lines)d\n'\
+            % self.__dict__
+
+
+class BaseVersionedFile(object):
+    """VersionedFile skeleton for MultiParent"""
+
+    def __init__(self, snapshot_interval=25, max_snapshots=None):
+        self._lines = {}
+        self._parents = {}
+        self._snapshots = set()
+        self.snapshot_interval = snapshot_interval
+        self.max_snapshots = max_snapshots
+
+    def versions(self):
+        return iter(self._parents)
+
+    def do_snapshot(self, version_id, parent_ids):
+        if self.snapshot_interval is None:
+            return False
+        if self.max_snapshots is not None and\
+            len(self._snapshots) == self.max_snapshots:
+            return False
+        if len(parent_ids) == 0:
+            return True
+        for ignored in xrange(self.snapshot_interval):
+            if len(parent_ids) == 0:
+                return False
+            version_ids = parent_ids
+            parent_ids = []
+            for version_id in version_ids:
+                if version_id not in self._snapshots:
+                    parent_ids.extend(self._parents[version_id])
+        else:
+            return True
+
+    def add_version(self, lines, version_id, parent_ids,
+                    force_snapshot=None, single_parent=False):
+        if force_snapshot is None:
+            do_snapshot = self.do_snapshot(version_id, parent_ids)
+        else:
+            do_snapshot = force_snapshot
+        if do_snapshot:
+            self._snapshots.add(version_id)
+            diff = MultiParent([NewText(lines)])
+        else:
+            if single_parent:
+                parent_lines = self.get_line_list(parent_ids[:1])
+            else:
+                parent_lines = self.get_line_list(parent_ids)
+            diff = MultiParent.from_lines(lines, parent_lines)
+            if diff.is_snapshot():
+                self._snapshots.add(version_id)
+        self.add_diff(diff, version_id, parent_ids)
+        self._lines[version_id] = lines
+
+    def get_parents(self, version_id):
+        return self._parents[version_id]
+
+    def make_snapshot(self, version_id):
+        snapdiff = MultiParent([NewText(self.cache_version(version_id))])
+        self.add_diff(snapdiff, version_id, self._parents[version_id])
+        self._snapshots.add(version_id)
+
+    def import_versionedfile(self, vf, snapshots, no_cache=True,
+                             single_parent=False, verify=False):
+        """Import all revisions of a versionedfile
+
+        :param vf: The versionedfile to import
+        :param snapshots: If provided, the revisions to make snapshots of.
+            Otherwise, this will be auto-determined
+        :param no_cache: If true, clear the cache after every add.
+        :param single_parent: If true, omit all but one parent text, (but
+            retain parent metadata).
+        """
+        assert no_cache or not verify
+        revisions = set(vf.versions())
+        total = len(revisions)
+        pb = ui.ui_factory.nested_progress_bar()
+        try:
+            while len(revisions) > 0:
+                added = set()
+                for revision in revisions:
+                    parents = vf.get_parents(revision)
+                    if [p for p in parents if p not in self._parents] != []:
+                        continue
+                    lines = [a + ' ' + l for a, l in
+                             vf.annotate_iter(revision)]
+                    if snapshots is None:
+                        force_snapshot = None
+                    else:
+                        force_snapshot = (revision in snapshots)
+                    self.add_version(lines, revision, parents, force_snapshot,
+                                     single_parent)
+                    added.add(revision)
+                    if no_cache:
+                        self.clear_cache()
+                        vf.clear_cache()
+                        if verify:
+                            assert lines == self.get_line_list([revision])[0]
+                            self.clear_cache()
+                    pb.update('Importing revisions',
+                              (total - len(revisions)) + len(added), total)
+                revisions = [r for r in revisions if r not in added]
+        finally:
+            pb.finished()
+
+    def select_snapshots(self, vf):
+        build_ancestors = {}
+        descendants = {}
+        snapshots = set()
+        for version_id in topo_iter(vf):
+            potential_build_ancestors = set(vf.get_parents(version_id))
+            parents = vf.get_parents(version_id)
+            if len(parents) == 0:
+                snapshots.add(version_id)
+                build_ancestors[version_id] = set()
+            else:
+                for parent in vf.get_parents(version_id):
+                    potential_build_ancestors.update(build_ancestors[parent])
+                if len(potential_build_ancestors) > self.snapshot_interval:
+                    snapshots.add(version_id)
+                    build_ancestors[version_id] = set()
+                else:
+                    build_ancestors[version_id] = potential_build_ancestors
+        return snapshots
+
+    def select_by_size(self, num):
+        """Select snapshots for minimum output size"""
+        num -= len(self._snapshots)
+        new_snapshots = self.get_size_ranking()[-num:]
+        return [v for n, v in new_snapshots]
+
+    def get_size_ranking(self):
+        versions = []
+        new_snapshots = set()
+        for version_id in self.versions():
+            if version_id in self._snapshots:
+                continue
+            diff_len = self.get_diff(version_id).patch_len()
+            snapshot_len = MultiParent([NewText(
+                self.cache_version(version_id))]).patch_len()
+            versions.append((snapshot_len - diff_len, version_id))
+        versions.sort()
+        return versions
+        return [v for n, v in versions]
+
+    def import_diffs(self, vf):
+        for version_id in vf.versions():
+            self.add_diff(vf.get_diff(version_id), version_id,
+                          vf._parents[version_id])
+
+    def get_build_ranking(self):
+        could_avoid = {}
+        referenced_by = {}
+        for version_id in topo_iter(self):
+            could_avoid[version_id] = set()
+            if version_id not in self._snapshots:
+                for parent_id in self._parents[version_id]:
+                    could_avoid[version_id].update(could_avoid[parent_id])
+                could_avoid[version_id].update(self._parents)
+                could_avoid[version_id].discard(version_id)
+            for avoid_id in could_avoid[version_id]:
+                referenced_by.setdefault(avoid_id, set()).add(version_id)
+        available_versions = list(self.versions())
+        ranking = []
+        while len(available_versions) > 0:
+            available_versions.sort(key=lambda x:
+                len(could_avoid[x]) *
+                len(referenced_by.get(x, [])))
+            selected = available_versions.pop()
+            ranking.append(selected)
+            for version_id in referenced_by[selected]:
+                could_avoid[version_id].difference_update(
+                    could_avoid[selected])
+            for version_id in could_avoid[selected]:
+                referenced_by[version_id].difference_update(
+                    referenced_by[selected]
+                )
+        return ranking
+
+    def clear_cache(self):
+        self._lines.clear()
+
+    def get_line_list(self, version_ids):
+        return [self.cache_version(v) for v in version_ids]
+
+    def cache_version(self, version_id):
+        try:
+            return self._lines[version_id]
+        except KeyError:
+            pass
+        diff = self.get_diff(version_id)
+        lines = []
+        reconstructor = _Reconstructor(self, self._lines,
+                                       self._parents)
+        reconstructor.reconstruct_version(lines, version_id)
+        self._lines[version_id] = lines
+        return lines
+
+
+class MultiMemoryVersionedFile(BaseVersionedFile):
+
+    def __init__(self, snapshot_interval=25, max_snapshots=None):
+        BaseVersionedFile.__init__(self, snapshot_interval, max_snapshots)
+        self._diffs = {}
+
+    def add_diff(self, diff, version_id, parent_ids):
+        self._diffs[version_id] = diff
+        self._parents[version_id] = parent_ids
+
+    def get_diff(self, version_id):
+        return self._diffs[version_id]
+
+    def destroy(self):
+        self._diffs = {}
+
+
+class MultiVersionedFile(BaseVersionedFile):
+
+    def __init__(self, filename, snapshot_interval=25, max_snapshots=None):
+        BaseVersionedFile.__init__(self, snapshot_interval, max_snapshots)
+        self._filename = filename
+        self._diff_offset = {}
+
+    def get_diff(self, version_id):
+        start, count = self._diff_offset[version_id]
+        infile = open(self._filename + '.mpknit', 'rb')
+        try:
+            infile.seek(start)
+            sio = StringIO(infile.read(count))
+        finally:
+            infile.close()
+        zip_file = GzipFile(None, mode='rb', fileobj=sio)
+        try:
+            file_version_id = zip_file.readline()
+            return MultiParent.from_patch(zip_file.readlines())
+        finally:
+            zip_file.close()
+
+    def add_diff(self, diff, version_id, parent_ids):
+        outfile = open(self._filename + '.mpknit', 'ab')
+        try:
+            start = outfile.tell()
+            try:
+                zipfile = GzipFile(None, mode='ab', fileobj=outfile)
+                zipfile.writelines(itertools.chain(
+                    ['version %s\n' % version_id], diff.to_patch()))
+            finally:
+                zipfile.close()
+            end = outfile.tell()
+        finally:
+            outfile.close()
+        self._diff_offset[version_id] = (start, end-start)
+        self._parents[version_id] = parent_ids
+
+    def destroy(self):
+        try:
+            os.unlink(self._filename + '.mpknit')
+        except OSError, e:
+            if e.errno != errno.ENOENT:
+                raise
+        try:
+            os.unlink(self._filename + '.mpidx')
+        except OSError, e:
+            if e.errno != errno.ENOENT:
+                raise
+
+    def save(self):
+        open(self._filename + '.mpidx', 'wb').write(bencode.bencode(
+            (self._parents, list(self._snapshots), self._diff_offset)))
+
+    def load(self):
+        self._parents, snapshots, self._diff_offset = bencode.bdecode(
+            open(self._filename + '.mpidx', 'rb').read())
+        self._snapshots = set(snapshots)
+
+
+class _Reconstructor(object):
+    """Build a text from the diffs, ancestry graph and cached lines"""
+
+    def __init__(self, diffs, lines, parents):
+        self.diffs = diffs
+        self.lines = lines
+        self.parents = parents
+        self.cursor = {}
+
+    def reconstruct(self, lines, parent_text, version_id):
+        """Append the lines referred to by a ParentText to lines"""
+        parent_id = self.parents[version_id][parent_text.parent]
+        end = parent_text.parent_pos + parent_text.num_lines
+        return self._reconstruct(lines, parent_id, parent_text.parent_pos,
+                                 end)
+
+    def _reconstruct(self, lines, req_version_id, req_start, req_end):
+        """Append lines for the requested version_id range"""
+        # stack of pending range requests
+        pending_reqs = [(req_version_id, req_start, req_end)]
+        while len(pending_reqs) > 0:
+            req_version_id, req_start, req_end = pending_reqs.pop()
+            # lazily allocate cursors for versions
+            try:
+                start, end, kind, data, iterator = self.cursor[req_version_id]
+            except KeyError:
+                iterator = self.diffs.get_diff(req_version_id).range_iterator()
+                start, end, kind, data = iterator.next()
+            if start > req_start:
+                iterator = self.diffs.get_diff(req_version_id).range_iterator()
+                start, end, kind, data = iterator.next()
+
+            # find the first hunk relevant to the request
+            while end <= req_start:
+                start, end, kind, data = iterator.next()
+            self.cursor[req_version_id] = start, end, kind, data, iterator
+            # if the hunk can't satisfy the whole request, split it in two,
+            # and leave the second half for later.
+            if req_end > end:
+                pending_reqs.append((req_version_id, end, req_end))
+                req_end = end
+            if kind == 'new':
+                lines.extend(data[req_start - start: (req_end - start)])
+            else:
+                # If the hunk is a ParentText, rewrite it as a range request
+                # for the parent, and make it the next pending request.
+                parent, parent_start, parent_end = data
+                new_version_id = self.parents[req_version_id][parent]
+                new_start = parent_start + req_start - start
+                new_end = parent_end + req_end - end
+                pending_reqs.append((new_version_id, new_start, new_end))
+
+    def reconstruct_version(self, lines, version_id):
+        length = self.diffs.get_diff(version_id).num_lines()
+        return self._reconstruct(lines, version_id, 0, length)
+
+def gzip_string(lines):
+    sio = StringIO()
+    data_file = GzipFile(None, mode='wb', fileobj=sio)
+    data_file.writelines(lines)
+    data_file.close()
+    return sio.getvalue()

=== added file 'test_multiparent.py'
--- test_multiparent.py	1970-01-01 00:00:00 +0000
+++ test_multiparent.py	2007-06-11 02:10:58 +0000
@@ -0,0 +1,213 @@
+from unittest import TestCase
+
+from bzrlib import tests
+
+import multiparent
+
+
+LINES_1 = "a\nb\nc\nd\ne\n".splitlines(True)
+LINES_2 = "a\nc\nd\ne\n".splitlines(True)
+LINES_3 = "a\nb\nc\nd\n".splitlines(True)
+
+
+class Mock(object):
+
+    def __init__(self, **kwargs):
+        self.__dict__ = kwargs
+
+
+class TestMulti(TestCase):
+
+    def test_compare_no_parent(self):
+        diff = multiparent.MultiParent.from_lines(LINES_1)
+        self.assertEqual([multiparent.NewText(LINES_1)], diff.hunks)
+
+    def test_compare_one_parent(self):
+        diff = multiparent.MultiParent.from_lines(LINES_1, [LINES_2])
+        self.assertEqual([multiparent.ParentText(0, 0, 0, 1),
+                          multiparent.NewText(['b\n']),
+                          multiparent.ParentText(0, 1, 2, 3)],
+                         diff.hunks)
+
+    def test_compare_two_parents(self):
+        diff = multiparent.MultiParent.from_lines(LINES_1, [LINES_2, LINES_3])
+        self.assertEqual([multiparent.ParentText(1, 0, 0, 4),
+                          multiparent.ParentText(0, 3, 4, 1)],
+                         diff.hunks)
+
+    def test_range_iterator(self):
+        diff = multiparent.MultiParent.from_lines(LINES_1, [LINES_2, LINES_3])
+        diff.hunks.append(multiparent.NewText(['q\n']))
+        self.assertEqual([(0, 4, 'parent', (1, 0, 4)),
+                          (4, 5, 'parent', (0, 3, 4)),
+                          (5, 6, 'new', ['q\n'])],
+                         list(diff.range_iterator()))
+
+    def test_eq(self):
+        diff = multiparent.MultiParent.from_lines(LINES_1)
+        diff2 = multiparent.MultiParent.from_lines(LINES_1)
+        self.assertEqual(diff, diff2)
+        diff3 = multiparent.MultiParent.from_lines(LINES_2)
+        self.assertFalse(diff == diff3)
+        self.assertFalse(diff == Mock(hunks=[multiparent.NewText(LINES_1)]))
+        self.assertEqual(multiparent.MultiParent(
+                         [multiparent.NewText(LINES_1),
+                          multiparent.ParentText(0, 1, 2, 3)]),
+                         multiparent.MultiParent(
+                         [multiparent.NewText(LINES_1),
+                          multiparent.ParentText(0, 1, 2, 3)]))
+
+    def test_to_patch(self):
+        self.assertEqual(['i 1\n', 'a\n', '\n', 'c 0 1 2 3\n'],
+            list(multiparent.MultiParent([multiparent.NewText(['a\n']),
+            multiparent.ParentText(0, 1, 2, 3)]).to_patch()))
+
+    def test_from_patch(self):
+        self.assertEqual(multiparent.MultiParent(
+            [multiparent.NewText(['a\n']),
+             multiparent.ParentText(0, 1, 2, 3)]),
+             multiparent.MultiParent.from_patch(
+             ['i 1\n', 'a\n', '\n', 'c 0 1 2 3\n']))
+        self.assertEqual(multiparent.MultiParent(
+            [multiparent.NewText(['a']),
+             multiparent.ParentText(0, 1, 2, 3)]),
+             multiparent.MultiParent.from_patch(
+             ['i 1\n', 'a\n', 'c 0 1 2 3\n']))
+
+    def test_num_lines(self):
+        mp = multiparent.MultiParent([multiparent.NewText(['a\n'])])
+        self.assertEqual(1, mp.num_lines())
+        mp.hunks.append(multiparent.NewText(['b\n', 'c\n']))
+        self.assertEqual(3, mp.num_lines())
+        mp.hunks.append(multiparent.ParentText(0, 0, 3, 2))
+        self.assertEqual(5, mp.num_lines())
+        mp.hunks.append(multiparent.NewText(['f\n', 'g\n']))
+        self.assertEqual(7, mp.num_lines())
+
+
+class TestNewText(TestCase):
+
+    def test_eq(self):
+        self.assertEqual(multiparent.NewText([]), multiparent.NewText([]))
+        self.assertFalse(multiparent.NewText(['a']) ==
+                         multiparent.NewText(['b']))
+        self.assertFalse(multiparent.NewText(['a']) == Mock(lines=['a']))
+
+    def test_to_patch(self):
+        self.assertEqual(['i 0\n', '\n'],
+                         list(multiparent.NewText([]).to_patch()))
+        self.assertEqual(['i 1\n', 'a', '\n'],
+                         list(multiparent.NewText(['a']).to_patch()))
+        self.assertEqual(['i 1\n', 'a\n', '\n'],
+                         list(multiparent.NewText(['a\n']).to_patch()))
+
+
+class TestParentText(TestCase):
+
+    def test_eq(self):
+        self.assertEqual(multiparent.ParentText(1, 2, 3, 4),
+                         multiparent.ParentText(1, 2, 3, 4))
+        self.assertFalse(multiparent.ParentText(1, 2, 3, 4) ==
+                         multiparent.ParentText(2, 2, 3, 4))
+        self.assertFalse(multiparent.ParentText(1, 2, 3, 4) ==
+                         Mock(parent=1, parent_pos=2, child_pos=3,
+                              num_lines=4))
+
+    def test_to_patch(self):
+        self.assertEqual(['c 0 1 2 3\n'],
+                         list(multiparent.ParentText(0, 1, 2, 3).to_patch()))
+
+
+REV_A = ['a\n', 'b\n', 'c\n', 'd\n']
+REV_B = ['a\n', 'c\n', 'd\n', 'e\n']
+REV_C = ['a\n', 'b\n', 'e\n', 'f\n']
+
+
+class TestVersionedFile(TestCase):
+
+    def add_version(self, vf, text, version_id, parent_ids):
+        vf.add_version([(t+'\n') for t in text], version_id, parent_ids)
+
+    def make_vf(self):
+        vf = multiparent.MultiMemoryVersionedFile()
+        self.add_version(vf, 'abcd', 'rev-a', [])
+        self.add_version(vf, 'acde', 'rev-b', [])
+        self.add_version(vf, 'abef', 'rev-c', ['rev-a', 'rev-b'])
+        return vf
+
+    def test_add_version(self):
+        vf = self.make_vf()
+        self.assertEqual(REV_A, vf._lines['rev-a'])
+        vf.clear_cache()
+        self.assertEqual(vf._lines, {})
+
+    def test_get_line_list(self):
+        vf = self.make_vf()
+        vf.clear_cache()
+        self.assertEqual(REV_A, vf.get_line_list(['rev-a'])[0])
+        self.assertEqual([REV_B, REV_C], vf.get_line_list(['rev-b', 'rev-c']))
+
+    @staticmethod
+    def reconstruct(vf, revision_id, start, end):
+        reconstructor = multiparent._Reconstructor(vf, vf._lines,
+                                                   vf._parents)
+        lines = []
+        reconstructor._reconstruct(lines, revision_id, start, end)
+        return lines
+
+    @staticmethod
+    def reconstruct_version(vf, revision_id):
+        reconstructor = multiparent._Reconstructor(vf, vf._lines,
+                                                   vf._parents)
+        lines = []
+        reconstructor.reconstruct_version(lines, revision_id)
+        return lines
+
+    def test_reconstructor(self):
+        vf = self.make_vf()
+        self.assertEqual(['a\n', 'b\n'], self.reconstruct(vf, 'rev-a',  0, 2))
+        self.assertEqual(['c\n', 'd\n'], self.reconstruct(vf, 'rev-a',  2, 4))
+        self.assertEqual(['e\n', 'f\n'], self.reconstruct(vf, 'rev-c',  2, 4))
+        self.assertEqual(['a\n', 'b\n', 'e\n', 'f\n'],
+                          self.reconstruct(vf, 'rev-c',  0, 4))
+        self.assertEqual(['a\n', 'b\n', 'e\n', 'f\n'],
+                          self.reconstruct_version(vf, 'rev-c'))
+
+    def test_reordered(self):
+        """Check for a corner case that requires re-starting the cursor"""
+        vf = multiparent.MultiMemoryVersionedFile()
+        # rev-b must have at least two hunks, so split a and b with c.
+        self.add_version(vf, 'c', 'rev-a', [])
+        self.add_version(vf, 'acb', 'rev-b', ['rev-a'])
+        # rev-c and rev-d must each have a line from a different rev-b hunk
+        self.add_version(vf, 'b', 'rev-c', ['rev-b'])
+        self.add_version(vf, 'a', 'rev-d', ['rev-b'])
+        # The lines from rev-c and rev-d must appear in the opposite order
+        self.add_version(vf, 'ba', 'rev-e', ['rev-c', 'rev-d'])
+        vf.clear_cache()
+        lines = vf.get_line_list(['rev-e'])[0]
+        self.assertEqual(['b\n', 'a\n'], lines)
+
+
+class TestMultiVersionedFile(tests.TestCaseInTempDir):
+
+    def test_save_load(self):
+        vf = multiparent.MultiVersionedFile('foop')
+        vf.add_version('a\nb\nc\nd'.splitlines(True), 'a', [])
+        vf.add_version('a\ne\nd\n'.splitlines(True), 'b', ['a'])
+        vf.save()
+        newvf = multiparent.MultiVersionedFile('foop')
+        newvf.load()
+        self.assertEqual('a\nb\nc\nd', ''.join(newvf.get_line_list(['a'])[0]))
+        self.assertEqual('a\ne\nd\n', ''.join(newvf.get_line_list(['b'])[0]))
+
+    def test_filenames(self):
+        vf = multiparent.MultiVersionedFile('foop')
+        vf.add_version('a\nb\nc\nd'.splitlines(True), 'a', [])
+        self.failUnlessExists('foop.mpknit')
+        self.failIfExists('foop.mpidx')
+        vf.save()
+        self.failUnlessExists('foop.mpidx')
+        vf.destroy()
+        self.failIfExists('foop.mpknit')
+        self.failIfExists('foop.mpidx')

